#!./bin/py

from ConfigParser import ConfigParser
from base64 import encodestring
from eea.bounce import detect
from eea.bounce.Bouncers import BouncerAPI
from imapclient import IMAPClient
from optparse import OptionParser
import email
import logging
import re
import xmlrpclib, httplib
import sys


try:
    from email.header import decode_header as decode
    decode
except ImportError:
    #python 2.4 compatibility
    from email.Header import decode_header as decode

try:
    from elementtree.ElementTree import ElementTree #, Element, tostring
    from elementtree.ElementTree import fromstring
    ElementTree, fromstring
except ImportError:
    #XXX: the code in this module has been tested only with elementtree; 
    #     using xml.etree may not actually work properly; xml.etree is only
    #     available in python >= 2.5, while elementtree needs to be installed on 
    #     python 2.4
    try:
        from xml.etree import ElementTree
        #Element = ElementTree.Element
        #tostring = ElementTree.tostring
        fromstring = ElementTree.fromstring
    except ImportError:
        raise Exception("You need to install elementtree or use Python >= 2.5")

logging.basicConfig(level=logging.DEBUG, filename="var/log/eea-bounce.log")
log = logging.getLogger("BounceCleanup")
log.addHandler(logging.StreamHandler(sys.stdout))

acre = re.compile(r'<(?P<addr>[^>]*)>:')

def read_settings(filepath):
    parser = ConfigParser()
    parser.read(filepath)
    return dict(parser.items('default'))


def authenticated(f):
    """Decorator that ensures that webcalls are always authenticated"""

    def wrap(self, *args, **kwds):
        self.login()
        return f(self, *args, **kwds)

    return wrap


def _flatten(msg, leaves):
    if msg.is_multipart():
        for part in msg.get_payload():
            _flatten(part, leaves)
    else:
        leaves.append(msg)


def flatten_msg(msg):
    """Completely transform a multipart message into text"""
    leaves = []
    _flatten(msg)
    texts = [p.get_payload() for p in leaves]
    return "\n".join(texts)


class Processor(object):
    """A bouncing emails processor
    """

    verp_format = None
    verp_re = r"(?P<newsid>.+)-%s-(?P<email_name>.+)=(?P<email_domain>.+)@(.*)"
    filters = []

    def __init__(self, options):
        self.options = options

    def _split_url(self, url):
        """Splits the url into a Base and a Method name

        The Method will be the last segment of the url. The base is the first
        segments, that is, all the rest that don't belong to Method
        """
        if url.endswith('/'):
            i = -2
        else:
            i = -1

        method = url.split("/")[i]
        base = "/".join(url.split("/")[:i])

        return base, method
        

    def _xmlrpc(self, url, params=None):
        """Makes an XMLRPC call to the specified url and returns the parametersi
        """

        url, method = self._split_url(url)

        server = xmlrpclib.Server(url, 
                  BasicAuthTransport(self.options.login_username,
                                     self.options.login_password,
                            ))

        logging.debug("Calling base [%s] with methods [%s] and params [%s]", url, method, params)
        if params:
            return getattr(server, method)(params)
        else:
            return getattr(server, method)()

    def get_run_parameters(self):
        """Connects to the web server and retrieves various running parameters"""
        o = self.options
        response = self._xmlrpc(o.config_url)

        etree = fromstring(response)
        print response
        newsletters = [e.text.strip() 
                        for e in etree.find('newsletters').findall('newsletter')]
        verp_prefix = etree.find("verp_prefix")

        #break down the filters into a list of re.compile'd lines of filters

        filters = etree.find("filters")

        if filters:
            filters = map(
                        lambda f:re.compile(f),
                        filter(             #remove lines starting with #
                               lambda l:l.startswith("#"),
                               filter(      #remove empty lines
                                     None,
                                     map(lambda line:line.strip(),
                                         filters.text.split('\n')
                                         )
                                     )
                                )
                        )

        return (filters, newsletters, verp_prefix)

    def process_bounced_subscribers(self, bounced):
        """Makes a call into the webserver announcing which addresses are found to be bouncing
        """

        text = "\n".join(["|".join([str(x), str(y)]) for x,y in bounced])
        self._xmlrpc(self.options.callback_url, text)

    def scan_message(self, msg):
        """Determines if a message is bounced, returns the newsletter id and email address

        First, it tries to extract email + newsletter id from a message

        Uses a regular expression to achieve this. Example as follows:
        >>> x = re.compile(r"(?P<newsid>.+)-no-reply-(?P<email_name>.+)=(?P<email_domain>.+)@(.*)")
        >>> z = x.match("some-news-id-no-reply-tiberiu.ichim=gmail.com@mail.eea.eu")
        >>> z.groups()
        ('some-news-id', 'tiberiu.ichim', 'gmail.com', 'mail.eea.eu')

        Return-Path:no-reply-newsletter-id-tiberiu.ichim=gmail.com@mail.eea.europa.eu
        """

        # 1. Determine if the message was bounced using VERP

        rp = msg.get('Return-Path')
        if rp:
            matches = self.verp_format.match(rp)
            if matches is not None:
                groups = matches.groups()
                newsletter_id = groups[0]
                email_address = "@".join([groups[1], groups[2]])

                log.debug("Bounced by VERP: %s: %s", newsletter_id, email_address)
                return newsletter_id, email_address

        # 2. Determine the bouncing address using the Mailman library
        addresses = BouncerAPI.ScanMessages(None, msg)
        if addresses is BouncerAPI.Stop:
            log.info("Message got a Stop from Mailman Scan Library, so it's not a bounce")
            return None
        if addresses:
            email_address = addresses[0]
            newsletter_id = self.newsletters[-1]

            log.debug("Bounced detected by Mailman Scan Library: %s: %s", newsletter_id, email_address)
            return newsletter_id, email_address

        # 3. Use the set of predetermined filters to determine if email is bouncing
        try:
            bd = detect.BounceParser(msg, self.filters)

            if bd.isNoUserFound():
                email_address = bd.email_address
                newsletter_id = self.newsletters[-1]

                log.debug('Bounced detected by custom filter')
                return newsletter_id, email_address

                #log.error("Message is bounced, but could not find bounce source")
                #return None
        except detect.BounceDetectionException:
            log.error("Extra bounce detection failed to parse message")

        log.debug("Message is not bouncing.")
        return None

    def run(self):
        filters, newsletters, verp_prefix = self.get_run_parameters()

        if not newsletters:
            print "The website reports no newsletters have been created. "\
                  "Cannot run without a valid list of sent newsletters"
            sys.exit(0)

        self.verp_format = re.compile(self.verp_re % verp_prefix)
        self.newsletters = newsletters
        self.filters = filters

        o = self.options
        bounced = []
        cleanup_uids = []

        server = IMAPClient(o.imap_server, use_uid=True, ssl=o.imap_use_ssl)
        server.login(o.imap_username, o.imap_password)
        server.select_folder(o.imap_folder)
        uids = server.search(['NOT DELETED'])

        for uid in uids:
            data = server.fetch(uid, ['RFC822'])[uid]['RFC822']
            msg = email.message_from_string(data)
            s = decode(msg['Subject'])[0]
            subject = s[0].decode(s[1] or "ascii")

            log.info(u"Processing %s, from %s to %s at %s",
                    subject, msg['From'], msg['To'], msg['Date'])

            bounced_info = self.scan_message(msg)
            if bounced_info:
                log.info("Message is bouncing %s for %s", 
                         bounced_info[0], bounced_info[1])
                        
                bounced.append(bounced_info)
                cleanup_uids.append(uid)


        if bounced:
            log.info("The following email addresses were found to be bouncing: %s", bounced)
            self.process_bounced_subscribers(bounced)

            log.debug("Moving processed messages from %s to %s", 
                        self.options.imap_folder, self.options.imap_backup_folder)
            server.copy(cleanup_uids, self.options.imap_backup_folder)
            server.delete_messages(cleanup_uids)
            server.expunge()

        sys.exit(0)


class BasicAuthTransport(xmlrpclib.Transport):

    _use_datetime = True
    verbose = False

    def __init__(self, username=None, password=None):
        self.username=username
        self.password=password

    def request(self, host, handler, request_body, verbose):
        # issue XML-RPC request

        h = httplib.HTTP(host)
        h.putrequest("POST", handler)

        # required by HTTP/1.1
        h.putheader("Host", host)

        # required by XML-RPC
        h.putheader("User-Agent", self.user_agent)
        h.putheader("Content-Type", "text/xml")
        h.putheader("Content-Length", str(len(request_body)))

        # basic auth
        if self.username is not None and self.password is not None:
            h.putheader("AUTHORIZATION", "Basic %s" % str.replace(
                    encodestring("%s:%s" % (self.username, self.password)),
                    "\012", ""))
        h.endheaders()

        if request_body:
            h.send(request_body)

        errcode, errmsg, headers = h.getreply()

        if errcode != 200:
            raise xmlrpclib.ProtocolError(
                host + handler,
                errcode, errmsg,
                headers
                )

        return self.parse_response(h.getfile()) 


def main():
    usage = "usage: %prog arg1 arg2"
    parser = OptionParser(usage=usage)
    parser.add_option("--plone_username", dest="login_username",
                        help="The username used to connect to Plone")

    parser.add_option("--plone_password", dest="login_password",
                        help="The password used to connect to Plone")

    parser.add_option("--login_url", dest="login_url", 
                        help="The location in the Plone site where the Login form is found")

    parser.add_option("--config_url", dest="config_url", 
                        help="The URL for the config page")

    parser.add_option("--callback_url", dest="callback_url", 
                        help="The URL for the action page")

    parser.add_option("--imap_username", dest="imap_username",
                        help="The username used to connect to the IMAP server")

    parser.add_option("--imap_password", dest="imap_password",
                        help="The password used to connect to the IMAP server")

    parser.add_option("--imap_server", dest="imap_server",
                        help="The address of the IMAP server")

    parser.add_option("--imap_folder", dest="imap_folder",
                        help="The name of the folder in the IMAP account where bounced messages are found")

    parser.add_option("--imap_backup_folder", dest="imap_backup_folder",
                        help="The name of the IMAP folder where processed messages will be moved")

    parser.add_option("--use_ssl", dest="imap_use_ssl", default=True,
                        help="Use SSL for IMAP connection?")

    options, _ignored = parser.parse_args()
    #log.debug("Got arguments %s", options)

    o = options

    if not (o.login_username and o.login_password and o.login_url and o.config_url and 
            o.callback_url and o.imap_username and o.imap_password and o.imap_server and 
            o.imap_folder and o.imap_backup_folder and o.imap_use_ssl):
        parser.print_help()
        sys.exit(0)

    try:
        p = Processor(options)
        p.run()
    except ImportError, e:
        print "An error occured while processing. You may have invalid parameters. See below for a list of arguments"
        print "The error was:", e
        print
        parser.print_help()
        sys.exit(0)


if __name__ == "__main__":
    main()
